"""
Utility functions
"""
import pint
from openscm_units import unit_registry


def convert_lambda_to_ecs(lambda_val, f2x=3.74 * unit_registry("W/m^2")):
    """
    Convert a lambda value to equilibrium climate sensitivity (ECS)

    Parameters
    ----------
    lambda_val : :obj:`pint.Quantity`
        Value of lambda to convert to ECS

    f2x : :obj:`pint.Quantity`
        Value of the forcing due to a doubling of atmospheric |CO2|
        to assume during the conversion

    Returns
    -------
    :obj:`pint.Quantity`
        ECS value

    Raises
    ------
    TypeError
        ``lambda_val`` or ``f2x`` is not a :obj:`pint.Quantity`.
    """
    if not isinstance(lambda_val, pint.Quantity):
        raise TypeError("lambda_val is not a pint.Quantity")

    if not isinstance(f2x, pint.Quantity):
        raise TypeError("f2x is not a pint.Quantity")

    return -f2x / lambda_val



"""
Physical constants used in calculations
"""
from openscm_units import unit_registry as ur

DENSITY_WATER = 1000 * ur("kg/m^3")
""":obj:`pint.Quantity` : density of water"""

HEAT_CAPACITY_WATER = 4181 * ur("J/delta_degC/kg")
""":obj:`pint.Quantity` : heat capacity of water"""



"""
Module containing the impulse response model

The 2-timescale impulse response model is mathematically equivalent to the
two-layer model without state dependence.
"""
import numpy as np
from openscm_units import unit_registry as ur

from .base import TwoLayerVariant, _calculate_geoffroy_helper_parameters
from .constants import DENSITY_WATER, HEAT_CAPACITY_WATER
from .errors import ModelStateError

# pylint: disable=invalid-name


class ImpulseResponseModel(
    TwoLayerVariant
):  # pylint: disable=too-many-instance-attributes
    """
    TODO: top line and paper references

    This implementation uses a forward-differencing approach. This means that
    temperature and ocean heat uptake values are start of timestep values. For
    example, temperature[i] is only affected by drivers from the i-1 timestep.
    In practice, this means that the first temperature and ocean heat uptake
    values will always be zero and the last value in the input drivers has no
    effect on model output.
    """

    _d1_unit = "yr"
    _d2_unit = "yr"
    _q1_unit = "delta_degC/(W/m^2)"
    _q2_unit = "delta_degC/(W/m^2)"
    _efficacy_unit = "dimensionless"
    _delta_t_unit = "yr"

    _erf_unit = "W/m^2"

    _temp1_unit = "delta_degC"
    _temp2_unit = "delta_degC"
    _rndt_unit = "W/m^2"

    _save_paras = (  # parameters to save when doing a run
        "d1",
        "d2",
        "q1",
        "q2",
        "efficacy",
    )

    _name = "two_timescale_impulse_response"  # model name

    def __init__(
        self,
        q1=0.3 * ur("delta_degC/(W/m^2)"),
        q2=0.4 * ur("delta_degC/(W/m^2)"),
        d1=9.0 * ur("yr"),
        d2=400.0 * ur("yr"),
        efficacy=1.0 * ur("dimensionless"),
        delta_t=1 / 12 * ur("yr"),
    ):  # pylint: disable=too-many-arguments
        """
        Initialise

        Raises
        ------
        ValueError
            d1 >= d2, d1 must be the short-timescale
        """
        self.q1 = q1
        self.q2 = q2
        self.d1 = d1
        self.d2 = d2
        self.efficacy = efficacy
        self.delta_t = delta_t

        if d1 >= d2:
            raise ValueError("The short-timescale must be d1")

        self._erf = np.zeros(1) * np.nan
        self._temp1_mag = np.zeros(1) * np.nan
        self._temp2_mag = np.zeros(1) * np.nan
        self._rndt_mag = np.zeros(1) * np.nan
        self._timestep_idx = np.nan

    @property
    def d1(self):
        """
        :obj:`pint.Quantity`
            Response timescale of first box
        """
        return self._d1

    @d1.setter
    def d1(self, val):
        self._assert_is_pint_quantity_with_units(val, "d1", self._d1_unit)
        self._d1 = val
        self._d1_mag = val.to(self._d1_unit).magnitude

    @property
    def d2(self):
        """
        :obj:`pint.Quantity`
            Response timescale of second box
        """
        return self._d2

    @d2.setter
    def d2(self, val):
        self._assert_is_pint_quantity_with_units(val, "d2", self._d2_unit)
        self._d2 = val
        self._d2_mag = val.to(self._d2_unit).magnitude

    @property
    def q1(self):
        """
        :obj:`pint.Quantity`
            Sensitivity of first box response to radiative forcing
        """
        return self._q1

    @q1.setter
    def q1(self, val):
        self._assert_is_pint_quantity_with_units(val, "q1", self._q1_unit)
        self._q1 = val
        self._q1_mag = val.to(self._q1_unit).magnitude

    @property
    def q2(self):
        """
        :obj:`pint.Quantity`
            Sensitivity of second box response to radiative forcing
        """
        return self._q2

    @q2.setter
    def q2(self, val):
        self._assert_is_pint_quantity_with_units(val, "q2", self._q2_unit)
        self._q2 = val
        self._q2_mag = val.to(self._q2_unit).magnitude

    @property
    def efficacy(self):
        """
        :obj:`pint.Quantity`
            Efficacy factor
        """
        return self._efficacy

    @efficacy.setter
    def efficacy(self, val):
        self._assert_is_pint_quantity_with_units(val, "efficacy", self._efficacy_unit)
        self._efficacy = val
        self._efficacy_mag = val.to(self._efficacy_unit).magnitude

    def _reset(self):
        if np.isnan(self.erf).any():
            raise ModelStateError(
                "The model's drivers have not been set yet, call "
                ":meth:`self.set_drivers` first."
            )

        self._timestep_idx = np.nan
        self._temp1_mag = np.zeros_like(self._erf_mag) * np.nan
        self._temp2_mag = np.zeros_like(self._erf_mag) * np.nan
        self._rndt_mag = np.zeros_like(self._erf_mag) * np.nan

    def _run(self):
        for _ in self.erf:
            self.step()

    def _step(self):
        if np.isnan(self._timestep_idx):
            self._timestep_idx = 0

        else:
            self._timestep_idx += 1

        if np.equal(self._timestep_idx, 0):
            self._temp1_mag[self._timestep_idx] = 0.0
            self._temp2_mag[self._timestep_idx] = 0.0
            self._rndt_mag[self._timestep_idx] = 0.0

        else:
            self._temp1_mag[self._timestep_idx] = self._calculate_next_temp(
                self._delta_t_mag,
                self._temp1_mag[self._timestep_idx - 1],
                self._q1_mag,
                self._d1_mag,
                self._erf_mag[self._timestep_idx - 1],
            )

            self._temp2_mag[self._timestep_idx] = self._calculate_next_temp(
                self._delta_t_mag,
                self._temp2_mag[self._timestep_idx - 1],
                self._q2_mag,
                self._d2_mag,
                self._erf_mag[self._timestep_idx - 1],
            )

            self._rndt_mag[self._timestep_idx] = self._calculate_next_rndt(
                self._temp1_mag[self._timestep_idx - 1],
                self._temp2_mag[self._timestep_idx - 1],
                self._erf_mag[self._timestep_idx - 1],
                self._efficacy_mag,
            )

    @staticmethod
    def _calculate_next_temp(delta_t, t, q, d, erf):
        decay_factor = np.exp(-delta_t / d)
        rise = erf * q * (1 - np.exp(-delta_t / d))

        return t * decay_factor + rise

    def _calculate_next_rndt(self, t1, t2, erf, efficacy):
        two_layer_paras = self.get_two_layer_parameters()
        lambda0 = two_layer_paras["lambda0"]

        if np.equal(efficacy, 1):
            efficacy_term = 0 * ur(self._erf_unit)
        else:
            gh = _calculate_geoffroy_helper_parameters(
                two_layer_paras["du"],
                two_layer_paras["dl"],
                two_layer_paras["lambda0"],
                two_layer_paras["efficacy"],
                two_layer_paras["eta"],
            )

            t1_h = t1 * ur(self._temp1_unit)
            t2_h = t2 * ur(self._temp2_unit)
            efficacy_term = (
                two_layer_paras["eta"]
                * (efficacy - 1)
                * ((1 - gh["phi1"]) * t1_h + (1 - gh["phi2"]) * t2_h)
            )

            if str(efficacy_term.units) != "watt / meter ** 2":
                raise AssertionError("units should have come out as W/m^2")

        out = erf - lambda0.magnitude * (t1 + t2) - efficacy_term.magnitude

        return out

    def _get_run_output_tss(self, ts_base):
        out_run_tss = []

        out_run_tss.append(
            self._create_ts(
                base=ts_base,
                unit=self._temp1_unit,
                variable="Surface Temperature|Box 1",
                values=self._temp1_mag,
            )
        )
        out_run_tss.append(
            self._create_ts(
                base=ts_base,
                unit=self._temp2_unit,
                variable="Surface Temperature|Box 2",
                values=self._temp2_mag,
            )
        )
        out_run_tss.append(
            self._create_ts(
                base=ts_base,
                unit=self._temp1_unit,
                variable="Surface Temperature",
                values=self._temp1_mag + self._temp2_mag,
            )
        )
        out_run_tss.append(
            self._create_ts(
                base=ts_base,
                unit=self._rndt_unit,
                variable="Heat Uptake",
                values=self._rndt_mag,
            )
        )

        return out_run_tss

    def get_two_layer_parameters(
        self,
    ):  # pylint:disable=missing-return-doc,missing-return-type-doc
        """
        Get equivalent two-layer model parameters

        For details on how the equivalence is calculated, please see the notebook
        ``impulse-response-equivalence.ipynb`` in the `OpenSCM Two Layer model
        repository <github.com/openscm/openscm-twolayermodel>`_.

        Returns
        -------
        dict of str : :obj:`pint.Quantity`
            Input arguments to initialise an
            :obj:`openscm_twolayermodel.TwoLayerModel` with the same
            temperature response as ``self``
        """
        lambda0 = 1 / (self.q1 + self.q2)
        C = (self.d1 * self.d2) / (self.q1 * self.d2 + self.q2 * self.d1)

        a1 = lambda0 * self.q1
        a2 = lambda0 * self.q2

        C_D = (lambda0 * (self.d1 * a1 + self.d2 * a2) - C) / self.efficacy
        eta = C_D / (self.d1 * a2 + self.d2 * a1)

        du = C / (DENSITY_WATER * HEAT_CAPACITY_WATER)
        dl = C_D / (DENSITY_WATER * HEAT_CAPACITY_WATER)

        out = {
            "lambda0": lambda0,
            "du": du,
            "dl": dl,
            "eta": eta,
            "efficacy": self.efficacy,
        }

        return out



"""
Module containing the base for model implementations
"""
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
import pint
import pint.errors
import tqdm.autonotebook as tqdman
from openscm_units import unit_registry as ur
from scmdata.run import ScmRun

from .constants import DENSITY_WATER, HEAT_CAPACITY_WATER
from .errors import UnitError

# pylint: disable=invalid-name


class Model(ABC):
    """
    Base class for model implementations
    """

    _save_paras = tuple()  # parameters to save when doing a run

    _name = None  # model name

    @staticmethod
    def _assert_is_pint_quantity_with_units(quantity, name, model_units):
        if not isinstance(quantity, pint.Quantity):
            raise TypeError("{} must be a pint.Quantity".format(name))

        try:
            quantity.to(model_units)
        except pint.errors.DimensionalityError as exc:
            raise UnitError("Wrong units for `{}`".format(name)) from exc

    @abstractmethod
    def set_drivers(self, *args, **kwargs):
        """
        Set the model's drivers
        """

    def reset(self):
        """
        Reset everything so that a new run can be performed.

        Called as late as possible before :meth:`run`.
        """
        self._reset()

    @abstractmethod
    def _reset(self):
        pass

    def run(self):
        """
        Run the model.
        """
        self._run()

    @abstractmethod
    def _run(self):
        pass

    def step(self):
        """
        Do a single time step.
        """
        self._step()

    @abstractmethod
    def _step(self):
        pass


class TwoLayerVariant(Model):
    """
    Base for variations of implementations of the two-layer model
    """

    _delta_t_unit = "s"
    _erf_unit = "W/m^2"

    @property
    def delta_t(self):
        """
        :obj:`pint.Quantity`
            Time step for forward-differencing approximation
        """
        return self._delta_t

    @delta_t.setter
    def delta_t(self, val):
        self._assert_is_pint_quantity_with_units(val, "delta_t", self._delta_t_unit)
        self._delta_t = val
        self._delta_t_mag = val.to(self._delta_t_unit).magnitude

    @property
    def erf(self):
        """
        :obj:`pint.Quantity`
            Effective radiative forcing
        """
        return self._erf

    @erf.setter
    def erf(self, val):
        self._assert_is_pint_quantity_with_units(val, "erf", self._erf_unit)
        self._erf = val
        self._erf_mag = val.to(self._erf_unit).magnitude

    def set_drivers(
        self, erf
    ):  # pylint: disable=arguments-differ # hmm need to think about this
        """
        Set drivers for a model run

        Parameters
        ----------
        erf : :obj:`pint.Quantity`
            Effective radiative forcing (W/m^2) to use to drive the model

        Raises
        ------
        AssertionError
            ``erf`` is not one-dimensional
        """
        if len(erf.shape) != 1:
            raise AssertionError("erf must be one-dimensional")

        self.erf = erf

    @staticmethod
    def _ensure_scenarios_are_scmrun(scenarios):
        if not isinstance(scenarios, ScmRun):
            driver = ScmRun(scenarios)
        else:
            driver = scenarios.copy()

        return driver

    @staticmethod
    def _create_ts(base, unit, variable, values):
        out = base.copy()
        out.index = out.index.set_levels([unit], "unit")
        out.index = out.index.set_levels([variable], "variable")
        out.iloc[:, :] = values

        return out

    @staticmethod
    def _select_timestep(driver):
        year_diff = driver["year"].diff().dropna()
        if (year_diff == 1).all():
            # assume yearly timesteps
            return 1 * ur("yr")

        time_diff = driver["time"].diff().dropna()
        if all(
            np.logical_and(
                time_diff <= np.timedelta64(31, "D"),
                time_diff >= np.timedelta64(28, "D"),
            )
        ):
            # Assume constant monthly timesteps. This is clearly an approximation but
            # while we have constant internal timesteps it's the best we can do.
            return 1 * ur("month")

        raise NotImplementedError(
            "Could not decide on timestep for time axis: {}".format(driver["time"])
        )

    def run_scenarios(  # pylint:disable=too-many-locals
        self, scenarios, driver_var="Effective Radiative Forcing", progress=True,
    ):
        """
        Run scenarios.

        The model timestep is automatically adjusted based on the timestep used in ``scenarios``.
        The timestep used in ``scenarios`` must be constant because this implementation
        has a constant timestep. Pull requests to upgrade the implementation to support
        variable timesteps are welcome `<https://github.com/openscm/openscm-twolayermodel/pulls>`_.

        Parameters
        ----------
        scenarios : :obj:`ScmDataFrame` or :obj:`ScmRun` or :obj:`pyam.IamDataFrame` or :obj:`pd.DataFrame` or :obj:`np.ndarray` or str
            Scenarios to run. The input will be converted to an :obj:`ScmRun` before
            the run takes place.

        driver_var : str
            The variable in ``scenarios`` to use as the driver of the model

        progress : bool
            Whether to display a progress bar

        Returns
        -------
        :obj:`ScmRun`
            Results of the run (including drivers)

        Raises
        ------
        ValueError
            No data is available for ``driver_var`` in the ``"World"`` region in
            ``scenarios``.
        """
        driver = self._ensure_scenarios_are_scmrun(scenarios)

        save_paras_meta = {
            "{} ({})".format(k, getattr(self, k).units): getattr(self, k).magnitude
            for k in self._save_paras
        }

        driver = driver.filter(variable=driver_var, region="World")
        if np.equal(driver.shape[0], 0):
            raise ValueError(
                "No World data available for driver_var `{}`".format(driver_var)
            )

        driver["climate_model"] = self._name
        for k, v in save_paras_meta.items():
            driver[k] = v

        timestep = self._select_timestep(driver)
        self.delta_t = timestep

        run_store = list()

        driver_ts = driver.timeseries()
        for i, (label, row) in tqdman.tqdm(
            enumerate(driver_ts.iterrows()),
            desc="scenarios",
            leave=False,
            disable=not (progress),
        ):
            meta = dict(zip(driver_ts.index.names, label))
            row_no_nan = row.dropna()

            self.set_drivers(row_no_nan.values * ur(meta["unit"]))
            self.reset()
            self.run()

            out_run_tss_base = row_no_nan.to_frame().T
            out_run_tss_base.index.names = driver_ts.index.names

            out_run_tss = [out_run_tss_base]
            out_run_tss += self._get_run_output_tss(out_run_tss[0])

            out_run = ScmRun(pd.concat(out_run_tss))
            out_run["run_idx"] = i

            run_store.append(out_run)

        idx = run_store[0].meta.columns.tolist()

        def get_ordered_timeseries(in_ts):
            in_ts = in_ts.reorder_levels(idx)

            return in_ts

        out = ScmRun(
            pd.concat(
                [get_ordered_timeseries(r.timeseries()) for r in run_store], axis=0
            )
        )

        return out

    @abstractmethod
    def _get_run_output_tss(self, ts_base):
        """Get the run output timeseries as a list"""


def _calculate_geoffroy_helper_parameters(  # pylint:disable=too-many-locals
    du, dl, lambda0, efficacy, eta
):
    C = du * HEAT_CAPACITY_WATER * DENSITY_WATER
    C_D = dl * HEAT_CAPACITY_WATER * DENSITY_WATER

    b_pt1 = (lambda0 + efficacy * eta) / (C)
    b_pt2 = (eta) / (C_D)
    b = b_pt1 + b_pt2
    b_star = b_pt1 - b_pt2
    delta = b ** 2 - (4 * lambda0 * eta) / (C * C_D)

    taucoeff = C * C_D / (2 * lambda0 * eta)
    tau1 = taucoeff * (b - delta ** 0.5)
    tau2 = taucoeff * (b + delta ** 0.5)

    phicoeff = C / (2 * efficacy * eta)
    phi1 = phicoeff * (b_star - delta ** 0.5)
    phi2 = phicoeff * (b_star + delta ** 0.5)

    adenom = C * (phi2 - phi1)
    a1 = tau1 * phi2 * lambda0 / adenom
    a2 = -tau2 * phi1 * lambda0 / adenom

    out = {
        "C": C,
        "C_D": C_D,
        "b": b,
        "b_star": b_star,
        "delta": delta,
        "tau1": tau1,
        "tau2": tau2,
        "phi1": phi1,
        "phi2": phi2,
        "a1": a1,
        "a2": a2,
    }

    return out



"""
OpenSCM Two Layer Model, implementations of the two layer radiative forcing driven models by `Held et al. <https://journals.ametsoc.org/doi/full/10.1175/2009JCLI3466.1>`_ and `Geoffroy et al. <https://journals.ametsoc.org/doi/pdf/10.1175/JCLI-D-12-00195.1>`_.

See README and docs for more info.
"""

from ._version import get_versions
from .impulse_response_model import ImpulseResponseModel  # noqa
from .two_layer_model import TwoLayerModel  # noqa

__version__ = get_versions()["version"]
del get_versions



"""
Module containing the two-layer model
"""
import numpy as np
from openscm_units import unit_registry as ur

from .base import TwoLayerVariant, _calculate_geoffroy_helper_parameters
from .constants import DENSITY_WATER, HEAT_CAPACITY_WATER
from .errors import ModelStateError

# pylint: disable=invalid-name


class TwoLayerModel(TwoLayerVariant):  # pylint: disable=too-many-instance-attributes
    """
    TODO: top line and paper references

    This implementation uses a forward-differencing approach. This means that
    temperature and ocean heat uptake values are start of timestep values. For
    example, temperature[i] is only affected by drivers from the i-1 timestep.
    In practice, this means that the first temperature and ocean heat uptake
    values will always be zero and the last value in the input drivers has no
    effect on model output.
    """

    _du_unit = "m"
    _heat_capacity_upper_unit = "J/delta_degC/m^2"
    _heat_capacity_lower_unit = "J/delta_degC/m^2"
    _dl_unit = "m"
    _lambda0_unit = "W/m^2/delta_degC"
    _a_unit = "W/m^2/delta_degC^2"
    _efficacy_unit = "dimensionless"
    _eta_unit = "W/m^2/delta_degC"
    _delta_t_unit = "s"

    _erf_unit = "W/m^2"

    _temp_upper_unit = "delta_degC"
    _temp_lower_unit = "delta_degC"
    _rndt_unit = "W/m^2"

    _save_paras = (  # parameters to save when doing a run
        "du",
        "dl",
        "lambda0",
        "a",
        "efficacy",
        "eta",
    )

    _name = "two_layer"  # model name

    def __init__(
        self,
        du=50 * ur("m"),
        dl=1200 * ur("m"),
        lambda0=3.74 / 3 * ur("W/m^2/delta_degC"),
        a=0.0 * ur("W/m^2/delta_degC^2"),
        efficacy=1.0 * ur("dimensionless"),
        eta=0.8 * ur("W/m^2/delta_degC"),
        delta_t=ur("yr").to("s"),
    ):  # pylint: disable=too-many-arguments
        """
        Initialise
        """
        self.du = du
        self.dl = dl
        self.lambda0 = lambda0
        self.a = a
        self.efficacy = efficacy
        self.eta = eta
        self.delta_t = delta_t

        self._erf = np.zeros(1) * np.nan
        self._temp_upper_mag = np.zeros(1) * np.nan
        self._temp_lower_mag = np.zeros(1) * np.nan
        self._rndt_mag = np.zeros(1) * np.nan
        self._timestep_idx = np.nan

    @property
    def du(self):
        """
        :obj:`pint.Quantity`
            Depth of upper layer
        """
        return self._du

    @du.setter
    def du(self, val):
        self._assert_is_pint_quantity_with_units(val, "du", self._du_unit)
        self._du = val
        self._du_mag = val.to(self._du_unit).magnitude
        self._heat_capacity_upper_mag = self.heat_capacity_upper.to(
            self._heat_capacity_upper_unit
        ).magnitude

    @property
    def heat_capacity_upper(self):
        """
        :obj:`pint.Quantity`
            Heat capacity of upper layer
        """
        return self.du * DENSITY_WATER * HEAT_CAPACITY_WATER

    @property
    def dl(self):
        """
        :obj:`pint.Quantity`
            Depth of lower layer
        """
        return self._dl

    @dl.setter
    def dl(self, val):
        self._assert_is_pint_quantity_with_units(val, "dl", self._dl_unit)
        self._dl = val
        self._dl_mag = val.to(self._dl_unit).magnitude
        self._heat_capacity_lower_mag = self.heat_capacity_lower.to(
            self._heat_capacity_lower_unit
        ).magnitude

    @property
    def heat_capacity_lower(self):
        """
        :obj:`pint.Quantity`
            Heat capacity of lower layer
        """
        return self.dl * DENSITY_WATER * HEAT_CAPACITY_WATER

    @property
    def lambda0(self):
        """
        :obj:`pint.Quantity`
            Initial climate feedback factor
        """
        return self._lambda0

    @lambda0.setter
    def lambda0(self, val):
        self._assert_is_pint_quantity_with_units(val, "lambda0", self._lambda0_unit)
        self._lambda0 = val
        self._lambda0_mag = val.to(self._lambda0_unit).magnitude

    @property
    def a(self):
        """
        :obj:`pint.Quantity`
            Dependence of climate feedback factor on temperature
        """
        return self._a

    @a.setter
    def a(self, val):
        self._assert_is_pint_quantity_with_units(val, "a", self._a_unit)
        self._a = val
        self._a_mag = val.to(self._a_unit).magnitude

    @property
    def efficacy(self):
        """
        :obj:`pint.Quantity`
            Efficacy factor
        """
        return self._efficacy

    @efficacy.setter
    def efficacy(self, val):
        self._assert_is_pint_quantity_with_units(val, "efficacy", self._efficacy_unit)
        self._efficacy = val
        self._efficacy_mag = val.to(self._efficacy_unit).magnitude

    @property
    def eta(self):
        """
        :obj:`pint.Quantity`
            Heat transport efficiency
        """
        return self._eta

    @eta.setter
    def eta(self, val):
        self._assert_is_pint_quantity_with_units(val, "eta", self._eta_unit)
        self._eta = val
        self._eta_mag = val.to(self._eta_unit).magnitude

    def _reset(self):
        if np.isnan(self.erf).any():
            raise ModelStateError(
                "The model's drivers have not been set yet, call "
                ":meth:`self.set_drivers` first."
            )

        self._timestep_idx = np.nan
        self._temp_upper_mag = np.zeros_like(self._erf_mag) * np.nan
        self._temp_lower_mag = np.zeros_like(self._erf_mag) * np.nan
        self._rndt_mag = np.zeros_like(self._erf_mag) * np.nan

    def _run(self):
        for _ in self.erf:
            self.step()

    def _step(self):
        if np.isnan(self._timestep_idx):
            self._timestep_idx = 0

        else:
            self._timestep_idx += 1

        if np.equal(self._timestep_idx, 0):
            self._temp_upper_mag[self._timestep_idx] = 0.0
            self._temp_lower_mag[self._timestep_idx] = 0.0
            self._rndt_mag[self._timestep_idx] = 0.0

        else:
            self._temp_upper_mag[self._timestep_idx] = self._calculate_next_temp_upper(
                self._delta_t_mag,
                self._temp_upper_mag[self._timestep_idx - 1],
                self._temp_lower_mag[self._timestep_idx - 1],
                self._erf_mag[self._timestep_idx - 1],
                self._lambda0_mag,
                self._a_mag,
                self._efficacy_mag,
                self._eta_mag,
                self._heat_capacity_upper_mag,
            )

            self._temp_lower_mag[self._timestep_idx] = self._calculate_next_temp_lower(
                self._delta_t_mag,
                self._temp_lower_mag[self._timestep_idx - 1],
                self._temp_upper_mag[self._timestep_idx - 1],
                self._eta_mag,
                self._heat_capacity_lower_mag,
            )

            self._rndt_mag[self._timestep_idx] = self._calculate_next_rndt(
                self._delta_t_mag,
                self._temp_lower_mag[self._timestep_idx],
                self._temp_lower_mag[self._timestep_idx - 1],
                self._heat_capacity_lower_mag,
                self._temp_upper_mag[self._timestep_idx],
                self._temp_upper_mag[self._timestep_idx - 1],
                self._heat_capacity_upper_mag,
            )

    @staticmethod
    def _calculate_next_temp_upper(  # pylint: disable=too-many-arguments
        delta_t, t_upper, t_lower, erf, lambda0, a, efficacy, eta, heat_capacity_upper
    ):
        lambda_now = lambda0 - a * t_upper
        heat_exchange = efficacy * eta * (t_upper - t_lower)
        dT_dt = (erf - lambda_now * t_upper - heat_exchange) / heat_capacity_upper

        return t_upper + delta_t * dT_dt

    @staticmethod
    def _calculate_next_temp_lower(
        delta_t, t_lower, t_upper, eta, heat_capacity_lower
    ):  # pylint: disable=too-many-arguments
        heat_exchange = eta * (t_upper - t_lower)
        dT_dt = heat_exchange / heat_capacity_lower

        return t_lower + delta_t * dT_dt

    @staticmethod
    def _calculate_next_rndt(  # pylint: disable=too-many-arguments
        delta_t,
        t_lower_now,
        t_lower_prev,
        heat_capacity_lower,
        t_upper_now,
        t_upper_prev,
        heat_capacity_upper,
    ):
        uptake_lower = heat_capacity_lower * (t_lower_now - t_lower_prev) / delta_t
        uptake_upper = heat_capacity_upper * (t_upper_now - t_upper_prev) / delta_t

        return uptake_upper + uptake_lower

    def _get_run_output_tss(self, ts_base):
        out_run_tss = []

        out_run_tss.append(
            self._create_ts(
                base=ts_base,
                unit=self._temp_upper_unit,
                variable="Surface Temperature|Upper",
                values=self._temp_upper_mag,
            )
        )
        out_run_tss.append(
            self._create_ts(
                base=ts_base,
                unit=self._temp_lower_unit,
                variable="Surface Temperature|Lower",
                values=self._temp_lower_mag,
            )
        )
        out_run_tss.append(
            self._create_ts(
                base=ts_base,
                unit=self._rndt_unit,
                variable="Heat Uptake",
                values=self._rndt_mag,
            )
        )

        return out_run_tss

    def get_impulse_response_parameters(self):  # pylint:disable=missing-return-doc
        """
        Get equivalent two-timescale impulse response model parameters

        For details on how the equivalence is calculated, please see the notebook
        ``impulse-response-equivalence.ipynb`` in the `OpenSCM Two Layer model
        repository <github.com/openscm/openscm-twolayermodel>`_.

        Returns
        -------
        dict of str : :obj:`pint.Quantity`
            Input arguments to initialise an
            :obj:`openscm_twolayermodel.ImpulseResponseModel` with the same
            temperature response as ``self``

        Raises
        ------
        ValueError
            ``self.a`` is non-zero, the two-timescale model does not support
            state-dependence.
        """
        if not np.equal(self.a.magnitude, 0):
            raise ValueError(
                "Cannot calculate impulse response parameters with "
                "non-zero a={}".format(self.a)
            )

        gh = _calculate_geoffroy_helper_parameters(
            self.du, self.dl, self.lambda0, self.efficacy, self.eta
        )

        d1 = gh["tau1"]
        d2 = gh["tau2"]

        qdenom = gh["C"] * (gh["phi2"] - gh["phi1"])
        q1 = gh["tau1"] * gh["phi2"] / qdenom
        q2 = -gh["tau2"] * gh["phi1"] / qdenom

        out = {
            "d1": d1,
            "d2": d2,
            "q1": q1,
            "q2": q2,
            "efficacy": self.efficacy,
        }

        return out



import versioneer
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

PACKAGE_NAME = "openscm-twolayermodel"
AUTHORS = [
    ("Zeb Nicholls", "zebedee.nicholls@climate-energy-college.org"),
    ("Jared Lewis", "jared.lewis@climate-energy-college.org"),
]
URL = "https://github.com/openscm/openscm-twolayermodel"

DESCRIPTION = (
    "Implementations of the two layer radiative forcing driven models by "
    "`Held et al. <https://journals.ametsoc.org/doi/full/10.1175/2009JCLI3466.1>`_ "
    "and `Geoffroy et al. <https://journals.ametsoc.org/doi/pdf/10.1175/JCLI-D-12-00195.1>`_"
)
LICENSE = "3-Clause BSD License"
KEYWORDS = [
    "openscm-twolayermodel",
    "openscm",
    "python",
    "repo",
    "simple",
    "climate",
    "model",
]
README = "README.rst"
CLASSIFIERS = [  # full list at https://pypi.org/pypi?%3Aaction=list_classifiers
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: BSD License",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]

SOURCE_DIR = "src"

REQUIREMENTS = ["scmdata>=0.9", "tqdm"]
REQUIREMENTS_NOTEBOOKS = [
    "ipywidgets",
    "notebook",
    "seaborn",
]
REQUIREMENTS_TESTS = [
    "codecov",
    "coverage",
    "nbval",
    "pytest-cov",
    "pytest>=4.0",
    "scipy",
]
REQUIREMENTS_DOCS = REQUIREMENTS_NOTEBOOKS + [
    "nbsphinx",
    "sphinx>=1.4",
    "sphinx_rtd_theme",
    "sphinx-click",
    "sphinx-copybutton",
]
REQUIREMENTS_DEPLOY = ["twine>=1.11.0", "setuptools>=38.6.0", "wheel>=0.31.0"]

REQUIREMENTS_DEV = [
    *[
        "bandit",
        "black==19.10b0",
        "black-nb",
        "flake8",
        "isort<5",  # isort 5 incompatible with pylint
        "mypy",
        "nbdime",
        "pydocstyle",
        "pylint>=2.4.4",
    ],
    *REQUIREMENTS_DEPLOY,
    *REQUIREMENTS_DOCS,
    *REQUIREMENTS_NOTEBOOKS,
    *REQUIREMENTS_TESTS,
]

REQUIREMENTS_EXTRAS = {
    "deploy": REQUIREMENTS_DEPLOY,
    "dev": REQUIREMENTS_DEV,
    "docs": REQUIREMENTS_DOCS,
    "notebooks": REQUIREMENTS_NOTEBOOKS,
    "tests": REQUIREMENTS_TESTS,
}

# Get the long description from the README file
with open(README, "r") as f:
    README_LINES = ["Openscm Two Layer Model", "=======================", ""]
    add_line = False
    for line in f:
        if line.strip() == ".. sec-begin-long-description":
            add_line = True
        elif line.strip() == ".. sec-end-long-description":
            break
        elif add_line:
            README_LINES.append(line.strip())

if len(README_LINES) < 3:
    raise RuntimeError("Insufficient description given")


class OpenscmTwoLayermodel(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({"test": OpenscmTwoLayermodel})

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description="\n".join(README_LINES),
    long_description_content_type="text/x-rst",
    author=", ".join([author[0] for author in AUTHORS]),
    author_email=", ".join([author[1] for author in AUTHORS]),
    url=URL,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=find_packages(SOURCE_DIR),  # no exclude as only searching in `src`
    package_dir={"": SOURCE_DIR},
    # include_package_data=True,
    install_requires=REQUIREMENTS,
    extras_require=REQUIREMENTS_EXTRAS,
    cmdclass=cmdclass,
)



"""
Exceptions raised within ``openscm_twolayermodel``
"""


class UnitError(ValueError):
    """
    Exception raised if something has the wrong units
    """


class ModelStateError(ValueError):
    """
    Exception raised if a model's state is incompatible with the action
    """
